#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json


class AcQuantumRawConfig:

    def __init__(self, system_config, one_q_gate_fidelities, qubit_parameter, system_status, two_q_gate_fidelity):
        # type: (dict, dict, dict, dict, dict) -> None
        self.system_config = BackendSystemConfig.from_dict(system_config)
        self.one_q_gate_fidelities = OneQGateFidelities.from_dict(one_q_gate_fidelities)
        self.qubit_parameter = QubitParameter.from_dict(qubit_parameter)
        self.system_status = SystemStatus.from_dict(system_status)
        self.two_q_gate_fidelity = TwoQGateFidelity.from_dict(two_q_gate_fidelity)

    @classmethod
    def from_json(cls, values):
        # type: ([dict]) -> AcQuantumRawConfig
        return AcQuantumRawConfig(*values)


class BackendSystemConfig:

    def __init__(self, computer_id, config_key, config_value):
        # type: (str, str, SystemConfigValue) -> None
        self.config_value = config_value  # type: SystemConfigValue
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> BackendSystemConfig
        config_value = SystemConfigValue.from_dict(json.loads(values['configValue']))
        return BackendSystemConfig(values['computerId'], values['configKey'], config_value)


class SystemConfigValue:

    def __init__(self, one_q_gates, one_q_gates_label, two_q_gates, two_q_gates_label, measure_size_upper_limit):
        # type: ([str], [str], [str], [str], int) -> None
        self.one_q_gates = one_q_gates
        self.one_q_gates_label = one_q_gates_label
        self.two_q_gates = two_q_gates
        self.two_q_gates_label = two_q_gates_label
        self.measure_size_upper_limit = measure_size_upper_limit

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> SystemConfigValue
        return SystemConfigValue(values['oneQGates'], values['oneQGatesLabel'], values['twoQGates'],
                                 values['twoQGatesLabel'], values['measureSizeUpperLimit'])


class OneQGateFidelities:

    def __init__(self, computer_id, config_key, config_value):
        self.config_value = config_value  # [dict]
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> OneQGateFidelities
        config_value = json.loads(values['configValue'])
        return OneQGateFidelities(values['computerId'], values['configKey'], config_value)


class QubitParameter:

    def __init__(self, computer_id, config_key, config_value):
        self.config_value = config_value
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> QubitParameter
        return QubitParameter(values['computerId'], values['configKey'], values['configValue'])


class SystemStatus:

    def __init__(self, computer_id, config_key, config_value):
        # type: (str, str, SystemStatusConfigValue) -> None
        self.config_value = config_value  # type: SystemStatusConfigValue
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> SystemStatus
        config_value = SystemStatusConfigValue.from_dict(json.loads(values['configValue']))
        return SystemStatus(values['computerId'], values['configKey'], config_value)


class SystemStatusConfigValue:

    def __init__(self, status, fridge_temperature, last_calibration_time):
        # type: (str, str, str) -> None
        self.last_calibration_time = last_calibration_time
        self.fridge_temperature = fridge_temperature
        self.status = status

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> SystemStatusConfigValue
        return SystemStatusConfigValue(values['status'], values['fridgeTemperature'], values['lastCalibrationTime'])


class TwoQGateFidelity:

    def __init__(self, computer_id, config_key, config_value):
        self.config_value = config_value  # type: [dict]
        self.config_key = config_key
        self.computer_id = computer_id

    @classmethod
    def from_dict(cls, values):
        # type: (dict) -> TwoQGateFidelity
        config_value = json.loads(values['configValue'])
        return TwoQGateFidelity(values['computerId'], values['configKey'], config_value)



#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import List, Union


class Gate(object):

    def __init__(self):
        # type: () -> None
        self.gateDetail = {}

    def set_gate_details(self, gate_details):
        # type: (dict) -> None
        self.gateDetail = gate_details


class HGate(Gate):

    def __init__(self, x, y):
        # type: (int, int) -> None
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(HGate, self).__init__()
        self.x = x
        self.y = y
        self.text = 'H'


class XGate(Gate):

    def __init__(self, x, y):
        # type: (int, int) -> None
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(XGate, self).__init__()
        self.x = x
        self.y = y
        self.text = 'X'


class YGate(Gate):

    def __init__(self, x, y):
        # type: (int, int) -> None
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(YGate, self).__init__()
        self.x = x
        self.y = y
        self.text = 'Y'


class ZGate(Gate):

    def __init__(self, x, y):
        # type: (int, int) -> None
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(ZGate, self).__init__()
        self.x = x
        self.y = y
        self.text = 'Z'


class SGate(Gate):

    def __init__(self, x, y):
        # type: (int, int) -> None
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(SGate, self).__init__()
        self.x = x
        self.y = y
        self.text = 'S'


class RxGate(Gate):

    def __init__(self, x, y, angle):
        # type: (int, int, int) -> None
        if angle not in range(0, 361):
            raise ValueError('Angle is not between 0 - 360')
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(RxGate, self).__init__()
        self.x = x
        self.y = y
        self.text = 'RX_{}'.format(angle)


class RyGate(Gate):

    def __init__(self, x, y, angle):
        # type: (int, int, int) -> None
        if angle not in range(0, 361):
            raise ValueError('Angle is not between 0 - 360')
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(RyGate, self).__init__()
        self.x = x
        self.y = y
        self.text = 'RY_{}'.format(angle)


class RzGate(Gate):

    def __init__(self, x, y, angle):
        # type: (int, int, int) -> None
        if angle not in range(0, 361):
            raise ValueError('Angle is not between 0 - 360')
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(RzGate, self).__init__()
        self.x = x
        self.y = y
        self.text = 'RZ_{}'.format(angle)


# PHASE
class CPhase(Gate):
    text = 'CP'

    def __init__(self, x, y):
        # type: (Union[List[int], int, (int, int)], Union[List[int], (int,int)]) -> None
        super(CPhase, self).__init__()
        if isinstance(x, int):
            self.x = self.x1 = x
        else:
            if len(set(x)) != 1:
                raise ValueError('X - coordinates not equal or empty')
            if len(x) == 1:
                self.x = self.x1 = x[0]
            else:
                if len(x) is not 2:
                    raise ValueError('X - coordinates greater 2')
                self.x, self.x1 = x

        if len(y) is not 2:
            raise ValueError('number of Y - coordinates unequal 2')
        if len(set(y)) is not 2:
            raise ValueError('Y - coordinates can not be equal')

        self.y, self.y1 = y
        self.text = 'CP'


class CCPhase(Gate):
    text = 'CCP'

    def __init__(self, x, y):
        # type: (Union[List[int], int, (int, int)], Union[List[int], (int,int)]) -> None
        super(CCPhase, self).__init__()
        if isinstance(x, int):
            if x <= 0:
                raise ValueError('x - coordinate must be greater 0')
            self.x = self.x1 = self.x2 = x
        else:
            if len(set(x)) is not 1:
                raise ValueError('X - coordinates not equal or empty')
            if not all(value > 0 for value in x):
                raise ValueError('X - coordinate must be greater 0')
            if len(x) == 1:
                self.x = x[0]
                self.x1 = x[0]
            else:
                if len(x) is not 3:
                    raise ValueError('X - coordinates greater 3')
                self.x, self.x1, self.x2 = x
        if len(y) is not 3:
            raise ValueError('number of Y - coordinates unequal 2')
        if len(set(y)) is not 3:
            raise ValueError('Y - coordinates can not be equal')
        if not all(value > 0 for value in y):
            raise ValueError('Y - coordinate must be greater 0')
        self.y, self.y1, self.y2 = y
        self.text = 'CCP'


class Measure(Gate):

    def __init__(self, x, y):
        # type: (int, int) -> None
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(Measure, self).__init__()
        self.x = x
        self.y = y
        self.text = 'M'


class SDag(Gate):

    def __init__(self, x, y):
        # type: (int, int) -> None
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(SDag, self).__init__()
        self.x = x
        self.y = y
        self.text = 'S†'


class TGate(Gate):

    def __init__(self, x, y):
        # type: (int, int) -> None
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(TGate, self).__init__()
        self.x = x
        self.y = y
        self.text = 'T'


class TDag(Gate):

    def __init__(self, x, y):
        # type: (int, int) -> None
        if x <= 0:
            raise ValueError('x - coordinate must be greater 0')
        if y <= 0:
            raise ValueError('y - coordinate must be greater 0')
        super(TDag, self).__init__()
        self.x = x
        self.y = y
        self.text = 'T†'



#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import json
import os
import re
from unittest import TestCase, mock

from acquantumconnector.connector.acquantumconnector import AcQuantumConnector
from acquantumconnector.credentials.credentials import AcQuantumCredentials
from acquantumconnector.model.backendtype import AcQuantumBackendType
from acquantumconnector.model.errors import AcQuantumRequestError
from acquantumconnector.model.gates import XGate


class MockApi:

    def __init__(self):
        self.cookies = None
        self.base_uri = 'http://quantumcomputer.ac.cn'
        self.failed_response = {
            'success': False,
            'exception': None,
            'message': 'error',
            'data': None
        }
        self.base_response = {
            'success': True,
            'exception': None,
            'message': None,
            'data': None
        }

    def get(self, *args, **kwargs):
        if args[0] == '{}/login'.format(self.base_uri):
            file_path = (os.path.dirname(__file__)) + '/resources/login.html'
            with open(file_path, 'r') as file:
                data = file.read().replace('\n', '')
            return MockedResponse({}, 200, data)
        elif args[0] == '{}/experiment/list'.format(self.base_uri):
            exp = [{
                'name': 'UnitTesting',
                'version': 1,
                'experimentId': 234,
                'type': 'SIMULATE',
                'execution': 0
            }, {
                'name': 'UnitTesting1',
                'version': 1,
                'experimentId': 235,
                'type': 'SIMULATE',
                'execution': 0
            }]
            body = self.base_response
            body['data'] = exp

            return MockedResponse(json_data=body)
        elif args[0] == '{}/experiment/detail'.format(self.base_uri):
            if 'params' in kwargs and kwargs['params']['experimentId'] == 123:
                body = self.base_response
                data = {
                    'code': '',
                    'version': 0,
                    'execution': 0,
                    'experimentName': 'UnitTesting',
                    'bitWidth': 10,
                    'experimentType': 'SIMULATE',
                    'data': [XGate(1, 1).__dict__]
                }
                body['data'] = data
                return MockedResponse(json_data=body)
            else:
                return MockedResponse(json_data=self.failed_response)
        elif args[0] == '{}/computerConfig/query'.format(self.base_uri):
            file_path = (os.path.dirname(__file__)) + '/resources/computer-config.json'
            with open(file_path) as f:
                data = json.load(f)
                return MockedResponse(json_data=data)
        elif args[0] == '{}/experiment/resultlist'.format(self.base_uri):
            result = [
                {'startTime': '2019-01-29 17:30:56',
                 'finishTime': '2019-01-29 17:30:57',
                 'process': None,
                 'data': {'1': '1.0'},
                 'id': 9868,
                 'seed': 0,
                 'shots': 100,
                 'measureQubits': [0]}]
            body = self.base_response
            if 'params' in kwargs and kwargs['params']['experimentId'] == 123:
                data = {
                    'realResult': [],
                    'simulateResult': result}
                body['data'] = data
                return MockedResponse(json_data=body)
            elif 'params' in kwargs and kwargs['params']['experimentId'] == 124:
                data = {
                    'realResult': result,
                    'simulateResult': []}
                body['data'] = data
                return MockedResponse(json_data=body)
            else:
                return MockedResponse(json_data=self.failed_response)
        else:
            return MockedResponse({'test': 'test'})

    def post(self, *args, **kwargs):
        if args[0] == '{}/login'.format(self.base_uri):
            if 'data' in kwargs and 'username' in kwargs['data']:
                return MockedResponse(json_data={'success': True})
            else:
                return MockedResponse({}, 403)
        elif args[0] == '{}/experiment/submit'.format(self.base_uri):
            if 'params' in kwargs and kwargs['params']['experimentId'] == 123:
                return MockedResponse(json_data=self.base_response)
        elif args[0] == '{}/experiment/delete'.format(self.base_uri):
            if 'params' in kwargs and kwargs['params']['experimentId'] == 123:
                return MockedResponse(json_data=self.base_response)
        elif args[0] == '{}/experiment/result/delete'.format(self.base_uri):
            if 'params' in kwargs and kwargs['params']['id'] == 123:
                return MockedResponse(json_data=self.base_response)

        return MockedResponse(json_data=self.failed_response)


class MockedResponse:

    def __init__(self, json_data=None, status_code=200, raw_response=None):
        self.json_data = json_data
        self.status_code = status_code
        self.raw_response = raw_response
        self.text = self.raw_response

    def json(self):
        return self.json_data


class AcQuantumConnectorUnitTest(TestCase):

    @classmethod
    def setUpClass(cls):
        cls.init()

    @classmethod
    @mock.patch('acquantumconnector.connector.acquantumconnector.requests.session', side_effect=MockApi)
    def init(cls, mock_api=None):
        cls.mock_api = mock_api
        cls.api = AcQuantumConnector()

    def setUp(self):
        self.api.create_session(AcQuantumCredentials(os.environ['ACQ_USER'], os.environ['ACQ_PWD']))

    def test_create_session(self):
        self.api.create_session(AcQuantumCredentials(os.environ['ACQ_USER'], os.environ['ACQ_PWD']))
        self.assertRegex(self.api._session.csrf, re.compile('[a-f\d]{8}(-[a-f\d]{4}){3}-[a-f\d]{12}?'))

    def test__request_csrf_token(self):
        token = self.api._request_csrf_token()
        self.assertRegex(token, re.compile('[a-f\d]{8}(-[a-f\d]{4}){3}-[a-f\d]{12}?'))

    def test_get_experiments(self):
        experiments = self.api.get_experiments()
        self.assertEqual(experiments[0].name, 'UnitTesting')
        self.assertEqual(experiments[1].name, 'UnitTesting1')

    def test_get_experiment(self):
        experiment = self.api.get_experiment(123)
        self.assertEqual('UnitTesting', experiment.detail.name)
        self.assertEqual('SIMULATE', experiment.detail.experiment_type)
        self.assertEqual(10, experiment.detail.bit_width)
        self.assertEqual('X', experiment.data[0]['text'])
        self.assertEqual(1, experiment.data[0]['x'])
        self.assertEqual(1, experiment.data[0]['y'])

    def test_get_experiment_should_fail(self):
        with self.assertRaises(AcQuantumRequestError):
            self.api.get_experiment(124)

    def test_delete_experiment(self):
        try:
            self.api.delete_experiment(123)
        except AcQuantumRequestError as e:
            self.fail(e)

    def test_delete_experiment_should_fail(self):
        with self.assertRaises(AcQuantumRequestError):
            self.api.delete_experiment(124)

    def test_get_backend_config(self):
        config = self.api.get_backend_config()
        status = config.system_status.config_value.status
        self.assertTrue(status in ['ONLINE', 'OFFLINE'])

    def test_run_experiment(self):
        self.api.run_experiment(123, AcQuantumBackendType.SIMULATE, bit_width=11, shots=100)

    def test_run_experiment_should_fail(self):
        with self.assertRaises(AcQuantumRequestError):
            self.api.run_experiment(124, AcQuantumBackendType.SIMULATE, bit_width=11, shots=100)

    def test_get_simulated_result(self):
        results = self.api.get_result(123)
        simulate_result = results.simulated_result[0]
        self.assertIsNotNone(simulate_result.start_time)
        self.assertIsNotNone(simulate_result.finish_time)
        self.assertIsNotNone(simulate_result.data)

    def test_delete_result(self):
        self.api.delete_result(123)

    def test_delete_result_should_fail(self):
        with self.assertRaises(AcQuantumRequestError):
            self.api.delete_result(124)



#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from typing import Any


class AcQuantumResponse(object):
    def __init__(self, success=True, data=None, exception=None):
        # type: (bool, Any, Any) -> None
        self.success = success
        self.data = data
        self.exception = exception

    def __str__(self):
        return 'AcResponse: {{ success: {}, exception: {} }}'.format(self.success, self.exception)

    def __repr__(self):
        return 'AcResponse: {{ success: {}, exception: {} }}'.format(self.success, self.exception)


class AcQuantumErrorResponse(AcQuantumResponse):

    def __init__(self, success=False, exception=None, status_code=None):
        # type: (bool, Any, int) -> None
        super(AcQuantumErrorResponse, self).__init__()
        self.status_code = status_code
        self.success = success
        self.exception = exception

    def __str__(self):
        return 'AcErrorResponse: {{ success: {}, exception: {}, status_code: {} }}'.format(self.success, self.exception,
                                                                                           self.status_code)

    def __repr__(self):
        return 'AcErrorResponse: {{ success: {}, exception: {}, status_code: {} }}'.format(self.success, self.exception,
                                                                                           self.status_code)


class AcQuantumExperimentDetail:

    def __init__(self, name, version, experiment_id, experiment_type, execution, bit_width=None):
        # type: (str, int, int, 'AcQuantumBackendType', int, int) -> None
        self.name = name
        self.version = version
        self.experiment_id = experiment_id
        self.experiment_type = experiment_type
        self.execution = execution
        self.bit_width = bit_width

    def __str__(self):
        return 'AcExperimentDetail: {{ name: {}, version: {}, experiment_id: {}, experiment_type: {}, execution: {}, ' \
               'bit_width {} }}' \
            .format(self.name, self.version, self.experiment_id, self.experiment_type, self.execution, self.bit_width)

    def __repr__(self):
        return 'AcExperimentDetail: {{ name: {}, version: {}, experiment_id: {}, experiment_type: {}, execution: {}, ' \
               'bit_width {} }}' \
            .format(self.name, self.version, self.experiment_id, self.experiment_type, self.execution, self.bit_width)


class AcQuantumExperiment:

    def __init__(self, detail, data, code=''):
        # type: (AcQuantumExperimentDetail, Any, str) -> None
        self.detail = detail
        self.data = data
        self.code = code

    def __str__(self):
        return 'AcExperiment: {{ data: {}, code: {}, detail: {} }}'.format(self.data, self.code, self.detail)

    def __repr__(self):
        return 'AcExperiment: {{ data: {}, code: {}, detail: {} }}'.format(self.data, self.code, self.detail)


class AcQuantumResultResponse:

    def __init__(self, simulated_result=None, real_result=None):
        # type: ([AcQuantumResult], [AcQuantumResult]) -> None
        self.simulated_result = simulated_result
        self.real_result = real_result

    def get_result(self):
        # type: () -> AcQuantumResult or (AcQuantumResult, AcQuantumResult)
        if self.simulated_result:
            if self.real_result:
                return self.real_result, self.simulated_result
            return self.simulated_result[0]
        elif self.real_result:
            return self.real_result[0]
        else:
            return None

    def __str__(self):
        return 'AcResultResponse {{ simulated_result: {}, real_result: {} }}'.format(self.simulated_result,
                                                                                     self.real_result)

    def __repr__(self):
        return 'AcResultResponse {{ simulated_result: {}, real_result: {} }}'.format(self.simulated_result,
                                                                                     self.real_result)


class AcQuantumResult:

    def __init__(self, result_id, seed, shots, start_time, measure_qubits, finish_time=None, process=None, data=None,
                 exception=None):
        # type: (int, int, int, str, [int], str, str, dict, Any) -> None
        self.result_id = result_id
        self.seed = seed
        self.shots = shots
        self.start_time = start_time
        self.measure_qubits = measure_qubits
        self.finish_time = finish_time
        self.process = process
        self.data = data
        self.exception = exception

    def __str__(self):
        return 'AcResult: {{ result_id: {}, seed: {}, shots: {}, start_time: {}, measure_qubits: {},  ' \
               'finish_time: {}, process: {}, data: {} }} '.format(self.result_id, self.seed, self.shots,
                                                                   self.start_time, self.measure_qubits,
                                                                   self.finish_time, self.process, self.data)

    def __repr__(self):
        return 'AcResult: {{ result_id: {}, seed: {}, shots: {}, start_time: {}, measure_qubits: {}, ' \
               'finish_time: {}, process: {}, data: {} }} '.format(self.result_id, self.seed, self.shots,
                                                                   self.start_time, self.measure_qubits,
                                                                   self.finish_time, self.process, self.data)



#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

from unittest import TestCase

from acquantumconnector.model.gates import RzGate, RxGate, RyGate, CPhase, CCPhase, XGate, YGate, ZGate, HGate, SGate, \
    TGate, SDag, TDag, Measure


class TestGates(TestCase):

    def setUp(self):
        pass

    def test_Rz_Gate(self):
        gate = RzGate(1, 1, 360)
        self.assertEqual(gate.text, 'RZ_360')
        self.assertEqual(gate.gateDetail, dict())
        self.assertEqual(gate.x, 1)
        self.assertEqual(gate.y, 1)
        g_dict = gate.__dict__
        self.assertTrue('x' in g_dict)
        self.assertTrue('y' in g_dict)
        self.assertTrue('gateDetail' in g_dict)
        self.assertTrue('text' in g_dict)

        with self.assertRaises(ValueError):
            RzGate(1, 1, 361)

        with self.assertRaises(ValueError):
            RzGate(1, 1, -1)

        with self.assertRaises(ValueError):
            RzGate(0, 1, 350)

        with self.assertRaises(ValueError):
            RzGate(1, 0, 234)

    def test_Rx_Gate(self):
        gate = RxGate(1, 1, 360)
        self.assertEqual(gate.text, 'RX_360')
        self.assertEqual(gate.gateDetail, dict())
        self.assertEqual(gate.x, 1)
        self.assertEqual(gate.y, 1)
        g_dict = gate.__dict__
        self.assertTrue('x' in g_dict)
        self.assertTrue('y' in g_dict)
        self.assertTrue('gateDetail' in g_dict)
        self.assertTrue('text' in g_dict)

        with self.assertRaises(ValueError):
            RxGate(1, 1, 361)

        with self.assertRaises(ValueError):
            RxGate(1, 1, -1)

        with self.assertRaises(ValueError):
            RxGate(0, 1, 350)

        with self.assertRaises(ValueError):
            RxGate(1, 0, 234)

    def test_Ry_Gate(self):
        gate = RyGate(1, 1, 360)
        self.assertEqual(gate.text, 'RY_360')
        self.assertEqual(gate.gateDetail, dict())
        self.assertEqual(gate.x, 1)
        self.assertEqual(gate.y, 1)
        g_dict = gate.__dict__
        self.assertTrue('x' in g_dict)
        self.assertTrue('y' in g_dict)
        self.assertTrue('gateDetail' in g_dict)
        self.assertTrue('text' in g_dict)

        with self.assertRaises(ValueError):
            RyGate(1, 1, 361)

        with self.assertRaises(ValueError):
            RyGate(1, 1, -1)

        with self.assertRaises(ValueError):
            RyGate(0, 1, 350)

        with self.assertRaises(ValueError):
            RyGate(1, 0, 234)

    def test_CPhase(self):
        phase = CPhase([1], [1, 2])
        self.assertEqual('CP', phase.text)
        self.assertEqual(dict(), phase.gateDetail)
        self.assertEqual(1, phase.x)
        self.assertEqual(1, phase.x1)
        self.assertEqual(1, phase.y)
        self.assertEqual(2, phase.y1)

        phase = CPhase(1, (1, 2))
        self.assertEqual(1, phase.x)
        self.assertEqual(1, phase.x1)
        self.assertEqual(1, phase.y)
        self.assertEqual(2, phase.y1)

        phase = CPhase((1, 1), (1, 2))
        self.assertEqual(1, phase.x)
        self.assertEqual(1, phase.x1)
        self.assertEqual(1, phase.y)
        self.assertEqual(2, phase.y1)

        with self.assertRaises(ValueError):
            CPhase([1, 2], [1, 2])

        with self.assertRaises(ValueError):
            CPhase([1, 1], [1, 2, 3])

        with self.assertRaises(ValueError):
            CPhase(1, (1, 2, 3))

        with self.assertRaises(ValueError):
            CPhase((1, 2), (1, 3))

        with self.assertRaises(ValueError):
            CPhase((1, 2), 1)

        with self.assertRaises(ValueError):
            CPhase(1, (3, 3))

        with self.assertRaises(ValueError):
            CPhase(1, [3, 3])

    def test_CCPhase(self):
        phase = CCPhase([1, 1, 1], [1, 2, 3])
        self.assertEqual('CCP', phase.text)
        self.assertEqual(dict(), phase.gateDetail)
        self.assertEqual(1, phase.x)
        self.assertEqual(1, phase.x1)
        self.assertEqual(1, phase.x2)
        self.assertEqual(1, phase.y)
        self.assertEqual(2, phase.y1)
        self.assertEqual(3, phase.y2)

        phase = CCPhase(1, (1, 2, 3))
        self.assertEqual(1, phase.x)
        self.assertEqual(1, phase.x1)
        self.assertEqual(1, phase.x2)
        self.assertEqual(1, phase.y)
        self.assertEqual(2, phase.y1)
        self.assertEqual(3, phase.y2)

        phase = CCPhase((1, 1, 1), (1, 2, 4))
        self.assertEqual(1, phase.x)
        self.assertEqual(1, phase.x1)
        self.assertEqual(1, phase.x2)
        self.assertEqual(1, phase.y)
        self.assertEqual(2, phase.y1)
        self.assertEqual(4, phase.y2)

        with self.assertRaises(ValueError):
            CCPhase([1, 2, 3], [1, 2, 3])

        with self.assertRaises(ValueError):
            CCPhase([1, 1, 1], [1, 2, 2])

        with self.assertRaises(ValueError):
            CCPhase(1, (1, 2, 2))

        with self.assertRaises(ValueError):
            CCPhase((1, 2), (1, 2, 3))

        with self.assertRaises(ValueError):
            CCPhase((1, 2), 1)

        with self.assertRaises(ValueError):
            CCPhase(1, (3, 3, 3))

    def test_XGate(self):
        gate = XGate(1, 1)
        self.assertEqual(1, gate.x)
        self.assertEqual(1, gate.y)
        self.assertEqual('X', gate.text)

        with self.assertRaises(ValueError):
            XGate(0, 1)

        with self.assertRaises(ValueError):
            XGate(1, 0)

    def test_YGate(self):
        gate = YGate(1, 1)
        self.assertEqual(1, gate.x)
        self.assertEqual(1, gate.y)
        self.assertEqual('Y', gate.text)

        with self.assertRaises(ValueError):
            YGate(0, 1)

        with self.assertRaises(ValueError):
            YGate(1, 0)

    def test_ZGate(self):
        gate = ZGate(1, 1)
        self.assertEqual(1, gate.x)
        self.assertEqual(1, gate.y)
        self.assertEqual('Z', gate.text)

        with self.assertRaises(ValueError):
            ZGate(0, 1)

        with self.assertRaises(ValueError):
            ZGate(1, 0)

    def test_HGate(self):
        gate = HGate(1, 1)
        self.assertEqual(1, gate.x)
        self.assertEqual(1, gate.y)
        self.assertEqual('H', gate.text)

        with self.assertRaises(ValueError):
            HGate(0, 1)

        with self.assertRaises(ValueError):
            HGate(1, 0)

    def test_SGate(self):
        gate = SGate(1, 1)
        self.assertEqual(1, gate.x)
        self.assertEqual(1, gate.y)
        self.assertEqual('S', gate.text)

        with self.assertRaises(ValueError):
            SGate(0, 1)

        with self.assertRaises(ValueError):
            SGate(1, 0)

    def test_TGate(self):
        gate = TGate(1, 1)
        self.assertEqual(1, gate.x)
        self.assertEqual(1, gate.y)
        self.assertEqual('T', gate.text)

        with self.assertRaises(ValueError):
            TGate(0, 1)

        with self.assertRaises(ValueError):
            TGate(1, 0)

    def test_SDag(self):
        gate = SDag(1, 1)
        self.assertEqual(1, gate.x)
        self.assertEqual(1, gate.y)
        self.assertEqual('S†', gate.text)

        with self.assertRaises(ValueError):
            SDag(0, 1)

        with self.assertRaises(ValueError):
            SDag(1, 0)

    def test_TDag(self):
        gate = TDag(1, 1)
        self.assertEqual(1, gate.x)
        self.assertEqual(1, gate.y)
        self.assertEqual('T†', gate.text)

        with self.assertRaises(ValueError):
            TDag(0, 1)

        with self.assertRaises(ValueError):
            TDag(1, 0)

    def test_Measure(self):
        measure = Measure(1, 1)
        self.assertEqual(1, measure.x)
        self.assertEqual(1, measure.y)
        self.assertEqual(1, measure.y)
        self.assertEqual('M', measure.text)

        with self.assertRaises(ValueError):
            Measure(0, 1)

        with self.assertRaises(ValueError):
            Measure(1, 0)



#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

class AcQuantumRequestError(Exception):
    def __init__(self, message, status_code=None):
        # type: (str, int) -> None
        self.status_code = status_code
        self.message = message

    def __str__(self):
        return 'AcRequestError: \'{}\''.format(self.message)

    def __repr__(self):
        return 'AcRequestError: \'{}\''.format(self.message)


class AcQuantumRequestForbiddenError(AcQuantumRequestError):
    def __init__(self, message='403 Forbidden'):
        # type: (str) -> None
        self.message = message

    def __str__(self):
        return '403 - AcRequestForbiddenError: \'{}\''.format(self.message)

    def __repr__(self):
        return '403 - AcRequestForbiddenError: \'{}\''.format(self.message)



from setuptools import setup

with open("acquantumconnector/_version.py") as f:
    version = f.readlines()[-1].split()[-1].strip("\"'")

with open('README.rst', 'r') as fh:
    long_description = fh.read()

with open('requirements.txt') as f:
    requirements = f.readlines()

info = {
    'name': 'acquantum-connector',
    'version': version,
    'author': 'Carsten Blank',
    'author_email': 'blank@data-cybernetics.com',
    'description': '',
    'long_description': long_description,
    'url': 'https://github.com/carstenblank/acquantum-connector',
    'packages': [
        'acquantumconnector',
        'acquantumconnector.connector',
        'acquantumconnector.credentials',
        'acquantumconnector.model',
    ],
    'install_requires': requirements,
    'license': 'Apache 2.0',
}

classifiers = [
    'Development Status :: 4 - Beta',
    'Programming Language :: Python :: 3',
    'Programming Language :: Python :: 3 :: Only',
    'License :: OSI Approved :: Apache Software License',
    'Operating System :: OS Independent',
]

setup(classifiers=classifiers, **info)



#  Copyright (c) 2019.  Carsten Blank
#
#   Licensed under the Apache License, Version 2.0 (the "License");
#   you may not use this file except in compliance with the License.
#   You may obtain a copy of the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in writing, software
#   distributed under the License is distributed on an "AS IS" BASIS,
#   WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#   See the License for the specific language governing permissions and
#   limitations under the License.

import pickle
import re
from typing import Any

import requests

from acquantumconnector.credentials.credentials import AcQuantumCredentials
from acquantumconnector.model.backendtype import AcQuantumBackendType
from acquantumconnector.model.config import AcQuantumRawConfig
from acquantumconnector.model.errors import AcQuantumRequestError, AcQuantumRequestForbiddenError
from acquantumconnector.model.gates import Gate
from acquantumconnector.model.response import AcQuantumExperimentDetail, AcQuantumExperiment, AcQuantumResult, \
    AcQuantumResultResponse, AcQuantumResponse


class AcQuantumSession(object):

    def __init__(self, csrf, cookies, credentials):
        # type: (str, Any, AcQuantumCredentials) -> None

        self.csrf = csrf
        self.cookies = cookies
        self.credentials = credentials

    def __str__(self):
        # type: () -> str
        return 'AcSession: [cookies: {}, csrf: {}]'.format(self.cookies, self.csrf)


class AcQuantumConnector(object):
    _schema = 'http'
    _base_uri = '{}://quantumcomputer.ac.cn'.format(_schema)
    _CHARSET_PARAM = ('_input_charset', 'utf-8')
    _TOKEN_HEADER_KEY = 'X-CSRF-TOKEN'

    def __init__(self):
        self._req = requests.session()
        self._credentials = None
        self._session = None

    def create_session(self, credentials):
        # type: (AcQuantumCredentials) -> None

        self._credentials = credentials
        csrf = self._request_csrf_token()
        cookies = self._req.cookies
        self._session = AcQuantumSession(csrf, cookies, credentials)
        response = self._login().json()
        if not response['success']:
            raise Exception('Connection refused: {}'.format(response['message']))

    def reconnect_session(self):
        print('... reconnecting session')
        self.create_session(self._credentials)

    def save_session(self):
        with open('session', 'wb') as f:
            self._session.cookies = self._req.cookies
            pickle.dump(self._session, f)

    def load_session(self):
        with open('session', 'rb') as f:
            self._session = pickle.load(f)
            self._req.cookies.update(self._session.cookies)

    def _login(self):
        # type: () -> requests.Response

        uri = '{}/login'.format(self._base_uri)
        params = {'_input_charset': 'utf-8'}
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        payload = {
            '_csrf': self._session.csrf,
            'username': self._credentials.user_name,
            'password': self._credentials.password,
        }

        return self._req.post(uri, data=payload, params=params, headers=headers)

    def create_experiment(self, bit_width, experiment_type, experiment_name):
        # type: (int, AcQuantumBackendType, str) -> int

        """
        :param bit_width:  bit width of the experiment
        :param experiment_type: Type of the backend the experiment should run
        :param experiment_name: Name of the experiment
        :return:
        """

        uri = '{}/experiment/infosave'.format(self._base_uri)
        params = {self._CHARSET_PARAM[0]: self._CHARSET_PARAM[1]}
        headers = {self._TOKEN_HEADER_KEY: self._session.csrf, 'Content-Type': 'application/json'}
        payload = {
            'bitWidth': bit_width,
            'type': experiment_type.name,
            'name': experiment_name
        }
        response = self.handle_ac_response(self._req.post(uri, params=params, headers=headers, json=payload))
        return response.data

    def update_experiment(self, experiment_id, gates, code=None, override=True):
        # type: (int, [Gate], str, bool) -> None

        """
        :param experiment_id: ID of created Experiment
        :param gates: Gates object definition that should be submitted
        :param code:
        :param override: Default = True. If False the last project State gets fetched from the Backend and Merged with
                the new Gates Definition
        :return: None
        :raises AcQuantumRequestError
        """

        uri = '{}/experiment/codesave'.format(self._base_uri)
        params = {
            self._CHARSET_PARAM[0]: self._CHARSET_PARAM[1]
        }
        headers = {self._TOKEN_HEADER_KEY: self._session.csrf, 'Content-Type': 'application/json'}
        payload = {
            'experimentId': str(experiment_id),
            'data': [gate.__dict__ for gate in gates],
            'code': code if code else ''
        }
        if not override:
            exp = self.get_experiment(experiment_id)
            existing_gates = exp.data
            payload['data'] = payload['data'] + existing_gates

        self.handle_ac_response(self._req.post(uri, json=payload, params=params, headers=headers))

    def get_experiment(self, experiment_id):
        # type: (int) -> AcQuantumExperiment
        """
        :param experiment_id: ID of experiment
        :return: AcQuantumExperiment
        """

        uri = '{}/experiment/detail'.format(self._base_uri)
        params = {
            'experimentId': experiment_id,
            self._CHARSET_PARAM[0]: self._CHARSET_PARAM[0]
        }
        headers = {self._TOKEN_HEADER_KEY: self._session.csrf}

        response = self.handle_ac_response(self._req.get(uri, params=params, headers=headers))
        body = response.data
        exp_detail = AcQuantumExperimentDetail(body['experimentName'], body['version'], int(experiment_id),
                                               body['experimentType'], body['execution'], bit_width=body['bitWidth'])
        experiment = AcQuantumExperiment(detail=exp_detail, data=body['data'], code=body['code'])
        return experiment

    def get_experiments(self):
        # type: () -> [AcQuantumExperimentDetail]

        uri = '{}/experiment/list'.format(self._base_uri)
        params = {
            self._CHARSET_PARAM[0]: self._CHARSET_PARAM[1]
        }
        headers = {self._TOKEN_HEADER_KEY: self._session.csrf, 'Content-Type': 'application/json'}

        response = self.handle_ac_response(self._req.get(uri, params=params, headers=headers))
        body = response.data
        experiment_list = [
            AcQuantumExperimentDetail(exp['name'], exp['version'], exp['experimentId'], exp['type'],
                                      exp['execution']) for exp in body]
        return experiment_list

    def run_experiment(self, experiment_id, experiment_type, bit_width, shots, seed=None):
        # type: (int, AcQuantumBackendType, int, int, str) -> None

        uri = '{}/experiment/submit'.format(self._base_uri)
        headers = {'Content-Type': 'application/json', self._TOKEN_HEADER_KEY: self._session.csrf}
        params = {
            self._CHARSET_PARAM[0]: self._CHARSET_PARAM[1],
            'experimentId': experiment_id,
            'bitWidth': bit_width,
            'type': experiment_type.name,
            'shots': shots,
            'seed': seed if seed else ''
        }
        self.handle_ac_response(self._req.post(uri, headers=headers, params=params))

    def get_result(self, experiment_id):
        # type: (int) -> AcQuantumResultResponse

        uri = '{}/experiment/resultlist'.format(self._base_uri)
        params = {
            'experimentId': experiment_id,
            self._CHARSET_PARAM[0]: self._CHARSET_PARAM[1]
        }
        headers = {self._TOKEN_HEADER_KEY: self._session.csrf}

        response = self.handle_ac_response(self._req.get(uri, params=params, headers=headers))
        body = response.data
        simulated_result = [
            AcQuantumResult(res['id'], res['seed'], res['shots'], res['startTime'], res['measureQubits'],
                            res['finishTime'],
                            res['process'], res['data']) for res in body['simulateResult']]
        real_result = [
            AcQuantumResult(res['id'], res['seed'], res['shots'], res['startTime'], res['measureQubits'],
                            res['finishTime'],
                            res['process'], res['data']) for res in body['realResult']]
        return AcQuantumResultResponse(simulated_result, real_result)

    def download_result(self, experiment_id, file_name=None):
        # type: (int, str) -> None

        uri = '{}/experiment/resultDownload'.format(self._base_uri)
        params = {'id': experiment_id}
        headers = {
            self._TOKEN_HEADER_KEY: self._session.csrf,
            'Accept-Encoding': 'gzip, deflate'
        }
        response = self._req.get(uri, params=params, headers=headers)
        if not file_name:
            file_name = 'result_{}.xls'.format(experiment_id)

        with open(file_name, 'wb') as f:
            f.write(response.content)

    def _request_csrf_token(self):
        # type: () -> str

        uri = '{}/login'.format(self._base_uri)
        response = self._req.get(uri)
        return re.search('[a-f\d]{8}(-[a-f\d]{4}){3}-[a-f\d]{12}?', response.text).group()

    def delete_experiment(self, experiment_id):
        # type: (int) -> None

        uri = '{}/experiment/delete'.format(self._base_uri)
        headers = {'Content-Type': 'application/json', self._TOKEN_HEADER_KEY: self._session.csrf}
        params = {
            self._CHARSET_PARAM[0]: self._CHARSET_PARAM[1],
            'experimentId': experiment_id,
        }

        self.handle_ac_response(self._req.post(uri, headers=headers, params=params))

    def delete_result(self, result_id):
        # type: (int) -> None
        uri = '{}/experiment/result/delete'.format(self._base_uri)
        headers = {'Content-Type': 'application/json', self._TOKEN_HEADER_KEY: self._session.csrf}
        params = {
            self._CHARSET_PARAM[0]: self._CHARSET_PARAM[1],
            'id': result_id
        }

        self.handle_ac_response(self._req.post(uri, headers=headers, params=params))

    def get_backend_config(self):
        # type: () -> AcQuantumRawConfig

        """

        :return: Backend Configuration
        """

        uri = '{}/computerConfig/query'.format(self._base_uri)
        headers = {'Content-Type': 'application/json', self._TOKEN_HEADER_KEY: self._session.csrf}
        params = {
            self._CHARSET_PARAM[0]: self._CHARSET_PARAM[1]
        }
        response = self.handle_ac_response(self._req.get(uri, headers=headers, params=params))
        return AcQuantumRawConfig.from_json(response.data)

    def available_backends(self):
        # TODO: implement
        # MOCKING
        return [
            {
                'backend_name': 'SIMULATE',
                'backend_version': '0.0.1',
                'n_qubits': 25,
                'basis_gates': ['XGate'],
                'gates': [
                    {'name': 'u1', 'parameters': ['lambda'], 'qasm_def': 'gate u1(lambda) q { U(0,0,lambda) q; }'},
                    {'name': 'u2', 'parameters': ['phi', 'lambda'],
                     'qasm_def': 'gate u2(phi,lambda) q { U(pi/2,phi,lambda) q; }'},
                    {'name': 'u3', 'parameters': ['theta', 'phi', 'lambda'],
                     'qasm_def': 'gate u3(theta,phi,lambda) q { U(theta,phi,lambda) q; }'},
                    {'name': 'cx', 'parameters': ['c', 't'], 'qasm_def': 'gate cx c,t { CX c,t; }'},
                    {'name': 'id', 'parameters': ['a'], 'qasm_def': 'gate id a { U(0,0,0) a; }'},
                    {'name': 'snapshot', 'parameters': ['slot'], 'qasm_def': 'gate snapshot(slot) q { TODO }'}
                ],
                'local': False,
                'simulator': True,
                'conditional': False,
                'open_pulse': False,
                'memory': False,
                'max_shots': 8192
            },
            {
                'backend_name': 'REAL',
                'backend_version': '0.0.1',
                'n_qubits': 11,
                'basis_gates': ['XGate'],
                'gates': [
                    {'name': 'x',
                     'parameters': [''],
                     'qasm_def': ''}
                ],
                'local': False,
                'simulator': False,
                'conditional': False,
                'open_pulse': False,
                'memory': False,
                'max_shots': 20000
            }
        ]

    @classmethod
    def handle_ac_response(cls, response):
        # type: (requests.Response) -> AcQuantumResponse

        r"""
        :param response: requests.Response
        :return: AcQuantumResponse
        :raises: AcRequestError, AcRequestForbiddenError
        """
        try:
            if response.status_code == 200:
                json = response.json()
                if json['success']:
                    try:
                        data = json['data']
                        return AcQuantumResponse(success=json['success'], exception=json['exception'], data=data)
                    except KeyError:
                        return AcQuantumResponse(success=json['success'], exception=json['exception'])
                else:
                    raise AcQuantumRequestError(json['exception'])
            else:
                error = response.json()
                if response.status_code == 403:
                    raise AcQuantumRequestForbiddenError()
                raise AcQuantumRequestError(error['exception'], status_code=response.status_code)
        except ValueError as err:
            raise AcQuantumRequestError(err.__str__())
